"""invrs_gym - A collection of inverse design challenges.

Copyright (c) 2023 The INVRS-IO authors.
"""

__version__ = "v0.2.1"
__author__ = "Martin F. Schubert <mfschubert@gmail.com>"

from invrs_gym import challenges as challenges
from invrs_gym import utils as utils
